######################################################################
package Swignition::MagicDuration;
######################################################################

use strict;
use overload '""'=>\&to_string, '+'=>\&add, '-'=>\&subtract, '<=>'=>\&compare, 'cmp'=>\&compare;
use Swignition::GenericParser::Utils;

sub new
{
	my $class        = shift;
	my $duration_obj = shift;
	my $this         = {};
	$this->{d}       = $duration_obj;
	
	bless $this, $class;
	return $this;
}

sub add
{
	my $this = shift;
	my $that = shift;
	my $sign = shift || '+';
	
	my $rv = $this->{d}->clone;
	if ($sign eq '-')
		{ $rv -= $that->{d}; }
	else
		{ $rv += $that->{d}; }
	
	return Swignition::MagicDuration::new($rv);
}

sub subtract
{
	return add(@_, '-');
}

sub compare
{
	my $this = shift;
	my $that = shift;
	return DateTime::Duration->compare($this->{d}, $that->{d}, DateTime->now);
}

sub parse
{
	my $class  = shift;
	my $string = shift;
	my $elem   = shift||undef;
	my $page   = shift||undef;
	
	# Try for nested class='s', class='min', class='h', etc. Standard=ISO-31.
	if ($elem)
	{
		my ($d, $h, $min, $s, $n);
		my $success = 0;
		my $X = {};
		
		# Find values.
		no strict;
		foreach my $x (qw(d h min s))
		{
			my @tmp = searchClass($x, $elem);
			if (@tmp)
			{
				my $y = STRINGIFY($tmp[0], {'abbr-pattern'=>1});
				$y    =~ s/\,/\./;
				$X->{$x} = "$y";  # MagicString -> string.
				$success++;
			}
		}
		
		if ($success)
		{
			# Cope with fractions.
			foreach my $frac (qw(d=24.h h=60.min min=60.s s=1000000000.n))
			{
				my ($big, $mult, $small) = split /[\=\.]/, $frac;
				next unless ($X->{$big} =~ /\./);
				
				my $int_part  = int($X->{$big});
				my $frac_part = $X->{$big} - $int_part;
				
				$X->{$big}    =  $int_part;
				$X->{$small} += ($mult * $frac_part);
			}
			use strict;
			$X->{'n'} = int($X->{'n'});
	
			# Construct and return object.
			my $dur = DateTime::Duration->new(
				days        => $X->{'d'}||0,
				hours       => $X->{'h'}||0,
				minutes     => $X->{'min'}||0,
				seconds     => $X->{'s'}||0,
				nanoseconds => $X->{'n'}||0
			);
			my $rv = Swignition::MagicDuration->new($dur);
			$rv->{string}  = $string;
			$rv->{element} = $elem;
			return $rv;
		}
	}

	# Commas as decimal points.
	my $string2 = $string;
	$string2 =~ s/\,/\./g;	
	
	# Standard=ISO-8601.
	if ($string2 =~ /^
			\s*
			([\+\-])?          # Potentially negitive...
			P                  # Period of...
			(?:([\d\.]*)Y)?    # n Years
			(?:([\d\.]*)M)?    # n Months
			(?:([\d\.]*)W)?    # n Weeks
			(?:([\d\.]*)D)?    # n Days
			(?:                 
				T               # And a time of...
				(?:([\d\.]*)H)? # n Hours
				(?:([\d\.]*)M)? # n Minutes
				(?:([\d\.]*)S)? # n Seconds
			)?
			\s*
			/ix)
	{
		my $X = {};
		$X->{'I'}   = $1;
		$X->{'y'}   = $2;
		$X->{'m'}   = $3;
		$X->{'w'}   = $4;
		$X->{'d'}   = $5;
		$X->{'h'}   = $6;
		$X->{'min'} = $7;
		$X->{'s'}   = $8;
		$X->{'n'}   = 0;
		
		# Handle fractional
		no strict;
		foreach my $frac (qw(y=12.m m=30.d w=7.d d=24.h h=60.min min=60.s s=1000000000.n))
		{
			my ($big, $mult, $small) = split /[\=\.]/, $frac;
			next unless ($X->{$big} =~ /\./);
			
			my $int_part  = int($X->{$big});
			my $frac_part = $X->{$big} - $int_part;
			
			$X->{$big}    =  $int_part;
			$X->{$small} += ($mult * $frac_part);
		}
		use strict;
		$X->{'n'} = int($X->{'n'});
		
		# Construct and return object.
		my $dur = DateTime::Duration->new(
			years       => $X->{'y'}||0,
			months      => $X->{'m'}||0,
			weeks       => $X->{'w'}||0,
			days        => $X->{'d'}||0,
			hours       => $X->{'h'}||0,
			minutes     => $X->{'min'}||0,
			seconds     => $X->{'s'}||0,
			nanoseconds => $X->{'n'}||0
		);
		my $rv = $X->{'I'} eq '-' ? Swignition::MagicDuration->new($dur->inverse) 
		                          : Swignition::MagicDuration->new($dur);
		$rv->{string}  = $string;
		$rv->{element} = $elem;
		return $rv;
	}
	
	# Duration as a simple number of seconds. Standard=SI.
	elsif ($string2 =~ /^\s* (\-?)(\d*)(?:\.(\d+))? \s* S? \s*$/ix && ($1||$2))
	{
		my $s = $2;
		my $n = "0.$3" * 1000000000;
		
		# Construct and return object.
		my $dur = DateTime::Duration->new(
			seconds     => $s,
			nanoseconds => $n
		);
		my $rv = $1 eq '-' ? Swignition::MagicDuration->new($dur->inverse) 
		                   : Swignition::MagicDuration->new($dur);
		$rv->{string}  = $string;
		$rv->{element} = $elem;
		return $rv;
	}

	# Look for hMeasure.
	elsif ($elem && $page)
	{
		# By this point, we're on a clone of the element, and certain class data
		# within it may have been destroyed. This is a little hack to find our
		# way back to the *real* element!
		my $real;
		my @real = $page->{DOM}->findnodes($elem->getAttribute('_xpath'));
		$real = $real[0] if (@real);
		return $string unless ($real);
		
		my @measures;
		if ($real->getAttribute('class') =~ /\b(hmeasure)\b/)
			{ push @measures, Swignition::uF::hMeasure::parse($page, $real); }
		else
			{ @measures = Swignition::uF::hMeasure::parse_all($page, $real); }
			
		foreach my $m (@measures)
		{
			next if ($m->{item});
			next if ($m->{type} && ($m->{type} !~ /^\s*(duration)\s*$/i));
			
			my ($dur, $neg);
			my $n = $m->{num};
			$n = "$n"; # MagicString -> string
			if ($n < 0)
			{
				$neg = 1;
				$n   = 0 - $n;
			}
			
			if ($m->{unit} && ($m->{unit} =~ /^\s* s ( ec (ond)? s? )? \s*$/ix))
			{
				print "hMeasure duration in seconds.\n";
				my $seconds     = int($n); $n -= $seconds; $n *= 1000000000;
				my $nanoseconds = int($n);
		
				# Construct and return object.
				$dur = DateTime::Duration->new(
					seconds     => $seconds,
					nanoseconds => $nanoseconds
				);
			}
			
			elsif ($m->{unit} && ($m->{unit} =~ /^\s* min ( (ute)? s? )? \s*$/ix))
			{
				print "hMeasure duration in minutes.\n";
				my $minutes     = int($n); $n -= $minutes; $n *= 60;
				my $seconds     = int($n); $n -= $seconds; $n *= 1000000000;
				my $nanoseconds = int($n);
		
				# Construct and return object.
				$dur = DateTime::Duration->new(
				   minutes     => $minutes,
					seconds     => $seconds,
					nanoseconds => $nanoseconds
				);
			}

			elsif ($m->{unit} && ($m->{unit} =~ /^\s* h ( our s? )? \s*$/ix))
			{
				print "hMeasure duration in hours.\n";
				my $hours       = int($n); $n -= $hours;   $n *= 60;
				my $minutes     = int($n); $n -= $minutes; $n *= 60;
				my $seconds     = int($n); 
		
				# Construct and return object.
				$dur = DateTime::Duration->new(
				   hours       => $hours,
				   minutes     => $minutes,
					seconds     => $seconds
				);
			}

			elsif ($m->{unit} && ($m->{unit} =~ /^\s* d ( ay s? )? \s*$/ix))
			{
				print "hMeasure duration in days.\n";
				my $days        = int($n); $n -= $days;    $n *= 24;
				my $hours       = int($n); $n -= $hours;   $n *= 60;
				my $minutes     = int($n); $n -= $minutes; $n *= 60;
				my $seconds     = int($n); 
		
				# Construct and return object.
				$dur = DateTime::Duration->new(
				   days        => $days,
				   hours       => $hours,
				   minutes     => $minutes,
					seconds     => $seconds
				);
			}
			
			if ($dur)
			{
				my $rv = ($neg==1) ? Swignition::MagicDuration->new($dur->inverse) 
		                         : Swignition::MagicDuration->new($dur);
				$rv->{string}   = $string;
				$rv->{element}  = $elem;
				$rv->{hmeasure} = $m;
				return $rv;
			}
		}
	}
	
	return $string;
}

sub to_string
{
	my $this = shift;
	my $str;
	
	# We coerce weeks into days and nanoseconds into fractions of a second
	# for compatibility with xsd:duration.
	
	if ($this->{d}->is_negative)
		{ $str .= '-P'; }
	else
		{ $str .= 'P'; }
		
	if ($this->{d}->years)
		{ $str .= $this->{d}->years.'Y'; }

	if ($this->{d}->months)
		{ $str .= $this->{d}->months.'M'; }

	if ($this->{d}->weeks || $this->{d}->days)
		{ $str .= ($this->{d}->days + (7 * $this->{d}->weeks)).'D'; }

	$str .= 'T';

	if ($this->{d}->hours)
		{ $str .= $this->{d}->hours.'H'; }

	if ($this->{d}->minutes)
		{ $str .= $this->{d}->minutes.'M'; }

	if ($this->{d}->seconds)
		{ $str .= ($this->{d}->seconds + ($this->{d}->nanoseconds / 1000000000)).'S'; }
		
	$str =~ s/T$//;
	
	return $str;
}

1;
